{% if box_content | trim | length %}
    {% set styles = [] %}
    {% if box_min_height %}
        {% do styles.append('min-height: ' ~ box_min_height ~ ';') %}
    {% endif %}
    {% if box_max_height %}
        {% do styles.append('max-height: ' ~ box_max_height ~ ';') %}
    {% endif %}
    {% if box_height %}
        {% do styles.append('height: ' ~ box_height ~ ';') %}
    {% endif %}
    {% if box_min_width %}
        {% do styles.append('min-width: ' ~ box_min_width ~ ';') %}
    {% endif %}
    {% if box_max_width %}
        {% do styles.append('max-width: ' ~ box_max_width ~ ';') %}
    {% endif %}
    {% if box_width %}
        {% do styles.append('width: ' ~ box_width ~ ';') %}
    {% endif %}
    <div{% if styles %} style="{{ styles | join }}"{% endif %}>
        {{ box_content }}
    </div>
{% endif %}
