{% extends 'base.html.j2' %}
{% set page_title = document.resource.label_plural | localize %}
{% do document.breadcrumbs.append(
    page_title,
    document.resource
) %}
{% block page_content %}
{% endblock %}
